namespace Geekeey.Common.Results.Tests;

[TestFixture]
internal sealed class ResultTransformTests
{
	[Test]
	public void Map_ReturnsSuccess_ForSuccess()
	{
		var start = Success(2);
		var result = start.Map(value => value.ToString());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public void Map_ReturnsFailure_ForFailure()
	{
		var start = Failure<int>("error");
		var result = start.Map(value => value.ToString());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void Then_ReturnsSuccess_ForSuccessAndMappingReturningSuccess()
	{
		var start = Success(2);
		var result = start.Then(value => Success(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public void Then_ReturnsFailure_ForSuccessAndMappingReturningFailure()
	{
		var start = Success(2);
		var result = start.Then(_ => Failure<string>("error"));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void Then_ReturnsFailure_ForFailureAndMappingReturningSuccess()
	{
		var start = Failure<int>("error");
		var result = start.Then(value => Success(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void Then_ReturnsFailure_ForFailureAndMappingReturningFailure()
	{
		var start = Failure<int>("error");
		var result = start.Then(_ => Failure<int>("error 2"));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void TryMap_ReturnsSuccess_ForSuccessWithoutThrowing()
	{
		var start = Success(2);
		var result = start.TryMap(value => value.ToString());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public void TryMap_ReturnsFailure_ForFailureWithoutThrowing()
	{
		var start = Failure<int>("error");
		var result = start.TryMap(value => value.ToString());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void TryMap_ReturnsFailure_ForSuccessWithThrowing()
	{
		var start = Success(2);
		var result = start.TryMap<string>(_ => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public void TryMap_ReturnsFailure_ForFailureWithThrowing()
	{
		var start = Failure<int>("error");
		var result = start.TryMap<string>(_ => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void ThenTry_ReturnsSuccess_ForSuccessAndMappingReturningSuccess()
	{
		var start = Success(2);
		var result = start.ThenTry(value => Success(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public void ThenTry_ReturnsFailure_ForSuccessAndMappingReturningFailure()
	{
		var start = Success(2);
		var result = start.ThenTry(_ => Failure<string>("error"));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void ThenTry_ReturnsFailure_ForFailureAndMappingReturningFailure()
	{
		var start = Failure<int>("error");
		var result = start.ThenTry(x => Success(x.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public void ThenTry_ReturnsFailure_ForSuccessAndMappingThrowing()
	{
		var start = Success(2);
		var result = start.ThenTry<string>(_ => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public void ThenTry_ReturnsFailure_ForFailureAndMappingThrowing()
	{
		var start = Failure<int>("error");
		var result = start.ThenTry<string>(_ => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task MapAsync_ReturnsSuccess_ForSuccess_Task()
	{
		var start = Success(2);
		var result = await start.MapAsync(value => Task.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task MapAsync_ReturnsFailure_ForFailure_Task()
	{
		var start = Failure<int>("error");
		var result = await start.MapAsync(value => Task.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenAsync_ReturnsSuccess_ForSuccessAndMappingReturningSuccess_Task()
	{
		var start = Success(2);
		var result = await start.ThenAsync(value => Task.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task ThenAsync_ReturnsFailure_ForSuccessAndMappingReturningFailure_Task()
	{
		var start = Success(2);
		var result = await start.ThenAsync(_ => Task.FromResult(Failure<string>("error")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenAsync_ReturnsFailure_ForFailureAndMappingReturningSuccess_Task()
	{
		var start = Failure<int>("error");
		var result = await start.ThenAsync(value => Task.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenAsync_ReturnsFailure_ForFailureAndMappingReturningFailure_Task()
	{
		var start = Failure<int>("error");
		var result = await start.ThenAsync(_ => Task.FromResult(Failure<int>("error 2")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsSuccess_ForSuccessWithoutThrowing_Task()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(value => Task.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsFailure_ForFailureWithoutThrowing_Task()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(value => Task.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsFailure_ForSuccessWithThrowing_Task()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(Task<string> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsFailure_ForSuccessWithAwaitThrowing_Task()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(async Task<string> (_) =>
		{
			await Task.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsFailure_ForFailureWithThrowing_Task()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(Task<string> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsFailure_ForFailureWithAwaitThrowing_Task()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(async Task<string> (_) =>
		{
			await Task.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsSuccess_ForSuccessAndMappingReturningSuccess_Task()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(value => Task.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForSuccessAndMappingReturningFailure_Task()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(_ => Task.FromResult(Failure<string>("error")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForFailureAndMappingReturningFailure_Task()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(x => Task.FromResult(Success(x.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForSuccessAndMappingThrowing_Task()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(Task<Result<string>> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForSuccessAndMappingAwaitThrowing_Task()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(async Task<Result<string>> (_) =>
		{
			await Task.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForFailureAndMappingThrowing_Task()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(Task<Result<string>> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForFailureAndMappingAwaitThrowing_Task()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(async Task<Result<string>> (_) =>
		{
			await Task.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task MapAsync_ReturnsSuccess_ForSuccess_ValueTask()
	{
		var start = Success(2);
		var result = await start.MapAsync(value => ValueTask.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task MapAsync_ReturnsFailure_ForFailure_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.MapAsync(value => ValueTask.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenAsync_ReturnsSuccess_ForSuccessAndMappingReturningSuccess_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenAsync(value => ValueTask.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task ThenAsync_ReturnsFailure_ForSuccessAndMappingReturningFailure_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenAsync(_ => ValueTask.FromResult(Failure<string>("error")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenAsync_ReturnsFailure_ForFailureAndMappingReturningSuccess_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.ThenAsync(value => ValueTask.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenAsync_ReturnsFailure_ForFailureAndMappingReturningFailure_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.ThenAsync(_ => ValueTask.FromResult(Failure<int>("error 2")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsSuccess_ForSuccessWithoutThrowing_ValueTask()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(value => ValueTask.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsFailure_ForFailureWithoutThrowing_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(value => ValueTask.FromResult(value.ToString()));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsFailure_ForSuccessWithThrowing_ValueTask()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(ValueTask<string> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsFailure_ForSuccessWithAwaitThrowing_ValueTask()
	{
		var start = Success(2);
		var result = await start.TryMapAsync(async ValueTask<string> (_) =>
		{
			await ValueTask.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsFailure_ForFailureWithThrowing_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(ValueTask<string> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task TryMapAsync_ReturnsFailure_ForFailureWithAwaitThrowing_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.TryMapAsync(async ValueTask<string> (_) =>
		{
			await ValueTask.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsSuccess_ForSuccessAndMappingReturningSuccess_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(value => ValueTask.FromResult(Success(value.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.Value, Is.EqualTo("2"));
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForSuccessAndMappingReturningFailure_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(_ => ValueTask.FromResult(Failure<string>("error")));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForFailureAndMappingReturningFailure_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(x => ValueTask.FromResult(Success(x.ToString())));

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForSuccessAndMappingThrowing_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(ValueTask<Result<string>> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForSuccessAndMappingAwaitThrowing_ValueTask()
	{
		var start = Success(2);
		var result = await start.ThenTryAsync(async ValueTask<Result<string>> (_) =>
		{
			await ValueTask.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<ExceptionError>());
			var instance = result.Error as ExceptionError;
			Assert.That(instance?.Exception, Is.InstanceOf<CustomTestException>());
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForFailureAndMappingThrowing_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(ValueTask<Result<string>> (_) => throw new CustomTestException());

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}

	[Test]
	public async Task ThenTryAsync_ReturnsFailure_ForFailureAndMappingAwaitThrowing_ValueTask()
	{
		var start = Failure<int>("error");
		var result = await start.ThenTryAsync(async ValueTask<Result<string>> (_) =>
		{
			await ValueTask.CompletedTask;
			throw new CustomTestException();
		});

		Assert.Multiple(() =>
		{
			Assert.That(result.IsSuccess, Is.False);
			Assert.That(result.Error, Is.InstanceOf<StringError>());
			Assert.That(result.Error?.Message, Is.EqualTo("error"));
		});
	}
}