﻿using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Diagnostics.Contracts;

namespace Geekeey.Common.Results;

/// <summary>
/// A type which contains either a success value or a failure value, which is represented by an <see cref="Error"/>.
/// </summary>
/// <typeparam name="T">The type of the success value.</typeparam>
[DebuggerTypeProxy(typeof(Result<>.ResultDebugProxy))]
public readonly partial struct Result<T>
{
	private readonly bool _success;

	private readonly T? _value;
	private readonly Error? _error;

	/// <summary>
	/// Creates a new result with an success value.
	/// </summary>
	/// <param name="value">The success value.</param>
	public Result(T value)
	{
		_success = true;
		_value = value;
		_error = default;
	}

	/// <summary>
	/// Creates a new result with an failure value.
	/// </summary>
	/// <param name="error">The error of the result.</param>
	public Result(Error error)
	{
		_success = false;
		_value = default;
		_error = error;
	}

	internal T? Value => _value;

	internal Error? Error => _success ? null : (_error ?? Error.DefaultValueError);

	/// <summary>
	/// Whether the result is a success.
	/// </summary>
	/// <remarks>
	/// This is always the inverse of <see cref="IsFailure"/> but is more specific about intent.
	/// </remarks>
	[MemberNotNullWhen(true, nameof(Value))]
	public bool IsSuccess => _success;

	/// <summary>
	/// Whether the result is a failure.
	/// </summary>
	/// <remarks>
	/// This is always the inverse of <see cref="IsSuccess"/> but is more specific about intent.
	/// </remarks>
	[MemberNotNullWhen(true, nameof(Error))]
	public bool IsFailure => !_success;

	/// <summary>
	/// Gets a string representation of the result.
	/// </summary>
	[Pure]
	public override string ToString()
		=> _success ? $"Success {{ {Value} }}" : $"Failure {{ {Error} }}";

	private sealed class ResultDebugProxy(Result<T> result)
	{
		public bool IsSuccess => result.IsSuccess;

		public object? Value => result.IsSuccess ? result.Value : result.Error;
	}
}